
prefixes = {
    '': 'https://openmbee.org/rdf/confluence-mdk/ontology/',
    'ac': 'https://openmbee.org/rdf/confluence-mdk/atlassian-confluence/',
    'dc': 'http://purl.org/dc/elements/1.1/',
    'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
    'owl': 'http://www.w3.org/2002/07/owl#',
    'xsd': 'http://www.w3.org/2001/XMLSchema#',
}
